# This code is taken from: https://github.com/python-trio/trio/pull/596/files
# Trio bills itself as; Trio – Pythonic async I/O for humans and snake people.
# Trio is released under a dual license of Apache or MIT
# 1. https://github.com/python-trio/trio/blob/master/LICENSE
# 2. https://github.com/python-trio/trio/blob/master/LICENSE.APACHE2
# 3. https://github.com/python-trio/trio/blob/master/LICENSE.MIT


import sys
import typing
import logging
import threading
import traceback


from . import logger

# TODO: This code is taken from a PR that has not been merged in Trio.
# We should monitor the tracking issue: https://github.com/python-trio/trio/issues/591
# Eventually we should adopt the solution that eventually gets merged into Trio(or alternatively; adopt Trio itself)


class BlockingTaskError(BlockingIOError):
    """
    Exception raised by wiji when the main asyncio thread is blocked by either
    an IO bound task execution or a CPU bound task execution.

    This Exception is raised in a separate thread(from the main asyncio thread) and thus does not
    impact the main thread.
    """

    pass


class BlockingWatchdog:
    """
    Monitors for any blocking calls in the main python asyncio thread.

    Python runs all asyncio operations(coroutines/tasks) on one Main thread in an evented manner.
    It is thus important that any operation you run in an asyncio environment be non-blocking.
    All your libraries(for http requests, database connections, rabbbitMQ clients etc) need to be async.
      As an example, the popular python http library/client; `python-requests <https://github.com/kennethreitz/requests>`_ is not async(non-blocking)
      For http requests, you should consider using an async client like `aiohttp <https://github.com/aio-libs/aiohttp>`_

    This class runs in a separate thread(away from the Main asyncio thread) so that it can monitor for any blocking calls on the Main thread.
    It does blocking detection in intervals of `watchdog_duration` seconds.
    The `watchdog_duration` is configurable and defaults to 0.1 seconds(0.1 seconds since that is also the default value in core Python.)
    We urge caution in trying to configure it to any value longer than 0.4 seconds;
      0.4s is the approximate `ping` duration(request & response) between the two farthest points on Earth for a signal travelling through copper.
      Of course, if you are doing any interstellar communication, then we urge you to consider using tools that are better suited for those
      kind of endeavours. `wiji` is not (as yet) suitable for interstellar IO communication.

    When `BlockingWatchdog` detects a blocking - IO/CPU bound - call that lasts for longer than `watchdog_duration` seconds;
    it will log an event that looks like:
        {
            "event": "wiji.BlockingWatchdog.blocked",
            "stage": "end",
            "error": "Blocked tasks Watchdog has not received any notifications in 0.1 seconds. This means the Main thread is blocked! "
            "Hint: are you running any tasks with blocking calls? eg; using python-requests? etc? "
            "Hint: look at the `stack_trace` attached to this log event to discover which calls are potentially blocking.",
            "stack_trace": [
                {
                    "thread_name": "MainThread",
                    "thread_stack_trace": [
                        "File cli/cli.py, line 269, in <module>\n    asyncio.run(async_main(), debug=True)\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/asyncio/runners.py, line 43, in run\n    return loop.run_until_complete(main)\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/asyncio/base_events.py, line 555, in run_until_complete\n    self.run_forever()\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/asyncio/base_events.py, line 523, in run_forever\n    self._run_once()\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/asyncio/base_events.py, line 1750, in _run_once\n    handle._run()\n",
                        "File /mystuff/wiji/wiji/worker.py, line 222, in consume_tasks\n    await self.run(*task_args, **task_kwargs)\n",
                        "File cli/cli.py, line 93, in run\n    resp = requests.get(url)\n",
                        "File /myVirtualenv/site-packages/requests/sessions.py, line 533, in request\n    resp = self.send(prep, **send_kwargs)\n",
                        "File /myVirtualenv/site-packages/requests/sessions.py, line 646, in send\n    r = adapter.send(request, **kwargs)\n",
                        "File /myVirtualenv/site-packages/requests/adapters.py, line 449, in send\n    timeout=timeout\n",
                        "File /myVirtualenv/site-packages/urllib3/connectionpool.py, line 600, in urlopen\n    chunked=chunked)\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/http/client.py, line 1321, in getresponse\n    response.begin()\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/socket.py, line 589, in readinto\n    return self._sock.recv_into(b)\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/ssl.py, line 1049, in recv_into\n    return self.read(nbytes, buffer)\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/ssl.py, line 908, in read\n    return self._sslobj.read(len, buffer)\n",
                    ],
                }
            ],
            "task_name": "_watchdogTask",
        }

    As you can see from that log event, you should as an application developer be able to identify where in your code the blocking calls are.
    When you do that, it is important you rectify them and make them async calls.
    If you do not, your task processing/execution is going to slow down considerably.

    If you buy into Python's asyncio world(which I think you should if your tasks/operations are mostly IO-bound), you should accept the fact that
    Python is running your stuff on one Thread.
    Your life will be much happier once you accept this and architect your operations with that in mind.
    """

    def __init__(self, watchdog_duration: float, task_name: str) -> None:
        if not isinstance(watchdog_duration, float):
            raise ValueError(
                """`watchdog_duration` should be of type:: `float` You entered: {0}""".format(
                    type(watchdog_duration)
                )
            )
        if not isinstance(task_name, str):
            raise ValueError(
                """`task_name` should be of type:: `str` You entered: {0}""".format(type(task_name))
            )

        self.watchdog_duration = watchdog_duration
        self.task_name = task_name

        self._stopped: bool = False
        self._thread: typing.Union[None, threading.Thread] = None
        self._notify_event: threading.Event = threading.Event()
        self._before_counter: int = 0
        self._after_counter: int = 0

        self.logger: logger.BaseLogger = logger.SimpleLogger("wiji.BlockingWatchdog")
        self.logger.bind(level="WARNING", log_metadata={"task_name": self.task_name})

    def notify_alive_before(self) -> None:
        """
        Notifies the watchdog that the Worker thread(the Event loop) is alive before running
        a task.
        """
        self._before_counter += 1
        self._notify_event.set()

    def notify_alive_after(self) -> None:
        """
        Notifies the watchdog that the Worker thread(the Event loop) is alive after running a
        task.
        """
        self._after_counter += 1

    def _main_loop(self) -> None:
        """
        Raises:
            BlockingTaskError: Exception raised when the main asyncio thread is blocked by either an IO/CPU bound task execution.
        """
        while True:
            if self._stopped:
                return

            self._notify_event.clear()
            orig_starts = self._before_counter
            orig_stops = self._after_counter

            if orig_starts == orig_stops:
                # main thread asleep; nothing to do until it wakes up
                self._notify_event.wait()
                if self._stopped:
                    return
            else:
                self._notify_event.wait(timeout=self.watchdog_duration)
                if self._stopped:
                    return

                if orig_starts == self._before_counter and orig_stops == self._after_counter:
                    try:
                        error_msg = (
                            "Blocked tasks Watchdog has not received any notifications in {watchdog_duration} seconds. "
                            "This means the Main thread is blocked! "
                            "\nHint: are you running any tasks with blocking calls? eg; using python-requests? etc? "
                            "\nHint: look at the `stack_trace` attached to this log event to discover which calls are potentially blocking.".format(
                                watchdog_duration=self.watchdog_duration
                            )
                        )
                        raise BlockingTaskError(error_msg)
                    except Exception as e:
                        all_threads_stack_trace = self._save_stack_trace()
                        # TODO: maybe we should log this at WARNING level, or DEBUG??
                        self.logger.log(
                            logging.ERROR,
                            {
                                "event": "wiji.BlockingWatchdog.blocked",
                                "stage": "end",
                                "error": str(e),
                                "stack_trace": all_threads_stack_trace,
                            },
                        )

    def _save_stack_trace(self) -> list:
        """
        This method returns a list of dictionaries each of which contains a thread name and its corresponding stack_trace.
        It returns something like:
            [
                {
                    "thread_name": "MainThread",
                    "thread_stack_trace": [
                        "File cli/cli.py, line 269, in <module>\n    asyncio.run(async_main(), debug=True)\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/asyncio/runners.py, line 43, in run\n    return loop.run_until_complete(main)\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/asyncio/base_events.py, line 555, in run_until_complete\n    self.run_forever()\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/asyncio/base_events.py, line 523, in run_forever\n    self._run_once()\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/asyncio/base_events.py, line 1750, in _run_once\n    handle._run()\n",
                        "File /mystuff/wiji/wiji/worker.py, line 222, in consume_tasks\n    await self.run(*task_args, **task_kwargs)\n",
                        "File cli/cli.py, line 93, in run\n    resp = requests.get(url)\n",
                        "File /myVirtualenv/site-packages/requests/sessions.py, line 533, in request\n    resp = self.send(prep, **send_kwargs)\n",
                        "File /myVirtualenv/site-packages/requests/sessions.py, line 646, in send\n    r = adapter.send(request, **kwargs)\n",
                        "File /myVirtualenv/site-packages/requests/adapters.py, line 449, in send\n    timeout=timeout\n",
                        "File /myVirtualenv/site-packages/urllib3/connectionpool.py, line 600, in urlopen\n    chunked=chunked)\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/http/client.py, line 1321, in getresponse\n    response.begin()\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/socket.py, line 589, in readinto\n    return self._sock.recv_into(b)\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/ssl.py, line 1049, in recv_into\n    return self.read(nbytes, buffer)\n",
                        "File /usr/python/3.7.0/3.7/lib/python3.7/ssl.py, line 908, in read\n    return self._sslobj.read(len, buffer)\n"
                    ]
                },
            ]
        From such a stack-trace, we can clearly see that we are using python requests to make network calls.
        And since python requests is not async, this is blocking the python event loop thread and thus slowing everything down.
        """
        # we could also use: faulthandler.dump_traceback(all_threads=True)
        stack_trace_of_all_threads_during_block = []
        for thread in threading.enumerate():
            thread_stack_trace = traceback.format_stack(
                f=sys._current_frames()[thread.ident]  # type: ignore
            )
            thread_name = thread.name
            if thread_name == "MainThread":
                # we are only interested in blocking ops in the asyncio thread
                stack_trace_of_all_threads_during_block.append(
                    {"thread_name": thread_name, "thread_stack_trace": thread_stack_trace}
                )
        return stack_trace_of_all_threads_during_block

    def start(self) -> None:
        self._thread = threading.Thread(
            target=self._main_loop, name="Thread-<wiji_watchdog>", daemon=True
        )
        self._thread.start()

    def stop(self) -> None:
        if typing.TYPE_CHECKING:
            # make mypy happy
            # https://github.com/python/mypy/issues/4805
            assert isinstance(self._thread, threading.Thread)
        self._stopped = True
        self._notify_event.set()
        self._thread.join()



about = {
    "__title__": "wiji",
    "__description__": "Wiji is an asyncio distributed task processor/queue.",
    "__url__": "https://github.com/komuw/wiji",
    "__version__": "v0.1.5",
    "__author__": "komuW",
    "__author_email__": "komuw05@gmail.com",
    "__license__": "MIT",
}



import abc
import enum
import uuid
import typing
import asyncio
import logging
import inspect

from . import hook
from . import logger
from . import broker
from . import protocol
from . import ratelimiter


# TODO: disambiguate which attributes should be in TaskOptions class
# and which ones should be in Task class.
# looks like the attributes that should be in TaskOptions class are the ones
# that have to do with the task as it is been called as opposed to when it is been declared.


class WijiMaxRetriesExceededError(Exception):
    """
    The tasks max_retries count has been exceeded.
    """

    pass


class WijiRetryError(Exception):
    """
    Exception that is raised so that `wiji.Worker` can know that current executing task is retrying.
    This enables `wiji.Worker` not to schedule any chained tasks of the current executing task.
    User applications should not capture this Exception!
    """

    pass


class TaskDelayError(Exception):
    """
    raised if `wiji` is unable to publish to the broker for any reason.
    """

    pass


@enum.unique
class TaskState(enum.Enum):
    QUEUEING: int = 1
    QUEUED: int = 2
    DEQUEUED: int = 3
    EXECUTING: int = 4
    EXECUTED: int = 5


class TaskOptions:
    def __init__(
        self,
        eta: float = 0.00,
        max_retries: int = 0,
        log_id: str = "",
        hook_metadata: typing.Union[None, str] = None,
        drain_duration: float = 10.0,
    ):
        self._validate_task_options_args(
            eta=eta,
            max_retries=max_retries,
            log_id=log_id,
            hook_metadata=hook_metadata,
            drain_duration=drain_duration,
        )
        self.eta = eta
        if self.eta < 0.00:
            self.eta = 0.00
        self.eta = protocol.Protocol._eta_to_isoformat(eta=self.eta)

        self.current_retries: int = 0
        self.max_retries = max_retries
        if self.max_retries < 0:
            self.max_retries = 0

        if log_id is not None:
            self.log_id = log_id
        else:
            self.log_id = ""

        if hook_metadata is not None:
            self.hook_metadata = hook_metadata
        else:
            self.hook_metadata = ""

        self.task_id: str = ""

        # `drain_duration` is the duration(in seconds) that a worker should wait
        # after getting a termination signal(SIGTERM, SIGQUIT etc).
        # during this duration, the worker does not consumer anymore tasks from the broker,
        # the worker will continue executing any tasks that it had already dequeued from the broker.
        # a simple way of choosing a value to set is:
        # drain_duration = time_taken_to_run_this_task + 1.00
        # eg: if your task is making a network call that lasts 30seconds,
        # thus; drain_duration = 30 + 1.00

        # the default value is 10.00 seconds.
        # mainly because that is also the default value of the process supervisor: `supervisord`
        self.drain_duration = drain_duration

        self.args: tuple = ()
        self.kwargs: dict = {}

    def __str__(self):
        return str(self.__dict__)

    def _validate_task_options_args(
        self,
        eta: float,
        max_retries: int,
        log_id: str,
        hook_metadata: typing.Union[None, str],
        drain_duration: float,
    ) -> None:
        if not isinstance(eta, float):
            raise ValueError(
                """`eta` should be of type:: `float` You entered: {0}""".format(type(eta))
            )
        if not isinstance(max_retries, int):
            raise ValueError(
                """`max_retries` should be of type:: `int` You entered: {0}""".format(
                    type(max_retries)
                )
            )
        if not isinstance(log_id, str):
            raise ValueError(
                """`log_id` should be of type:: `str` You entered: {0}""".format(type(log_id))
            )
        if not isinstance(hook_metadata, (type(None), str)):
            raise ValueError(
                """`hook_metadata` should be of type:: `None` or `str` You entered: {0}""".format(
                    type(hook_metadata)
                )
            )
        if not isinstance(drain_duration, float):
            raise ValueError(
                """`drain_duration` should be of type:: `float` You entered: {0}""".format(
                    type(drain_duration)
                )
            )


class Task(abc.ABC):
    """
    call it as:
        Task()(33,"hello", name="komu")

    usage:
        broker = wiji.broker.InMemoryBroker()
        task = Task(
                the_broker=broker,
                queue_name="PrintQueue",
            )
        task.delay(33, "hello", name="komu")

    You can also chain things as:
        task1 = wiji.task.Task()
        task2 = wiji.task.Task(chain=task1)
        task3 = wiji.task.Task(chain=task2)
    """

    def __init__(
        self,
        the_broker: broker.BaseBroker,
        queue_name: str,
        task_name: typing.Union[None, str] = None,
        chain: typing.Union[None, "Task"] = None,
        the_hook: typing.Union[None, hook.BaseHook] = None,
        the_ratelimiter: typing.Union[None, ratelimiter.BaseRateLimiter] = None,
        loglevel: str = "DEBUG",
        log_metadata: typing.Union[None, dict] = None,
        log_handler: typing.Union[None, logger.BaseLogger] = None,
    ) -> None:
        self._validate_task_args(
            the_broker=the_broker,
            queue_name=queue_name,
            task_name=task_name,
            chain=chain,
            the_hook=the_hook,
            the_ratelimiter=the_ratelimiter,
            loglevel=loglevel,
            log_metadata=log_metadata,
            log_handler=log_handler,
        )

        self.task_options = TaskOptions()
        self.the_broker = the_broker
        self.queue_name = queue_name
        self.chain = chain
        self.loglevel = loglevel.upper()

        if task_name is not None:
            self.task_name = task_name
        else:
            self.task_name = self.__class__.__name__

        if log_metadata is not None:
            self.log_metadata = log_metadata
        else:
            self.log_metadata = {}
        self.log_metadata.update({"task_name": self.task_name, "queue_name": self.queue_name})

        if log_handler is not None:
            self.logger = log_handler
        else:
            self.logger = logger.SimpleLogger(
                "wiji.Task.task_name={0}.task_id={1}".format(
                    self.task_name, self.task_options.task_id
                )
            )
        self.logger.bind(level=self.loglevel, log_metadata=self.log_metadata)
        self._sanity_check_logger(event="task_sanity_check_logger")

        if the_hook is not None:
            self.the_hook = the_hook
        else:
            self.the_hook = hook.SimpleHook(log_handler=self.logger)

        if the_ratelimiter is not None:
            self.the_ratelimiter = the_ratelimiter
        else:
            self.the_ratelimiter = ratelimiter.SimpleRateLimiter(log_handler=self.logger)

        self._checked_broker = False

    async def __call__(self, *args, **kwargs):
        await self.run(*args, **kwargs)

    def __str__(self):
        return str(
            {
                "task_name": self.task_name,
                "the_broker": self.the_broker,
                "queue_name": self.queue_name,
                "chain": self.chain,
                "task_options": self.task_options.__dict__,
            }
        )

    def _validate_task_args(
        self,
        the_broker: broker.BaseBroker,
        queue_name: str,
        task_name: typing.Union[None, str],
        chain: typing.Union[None, "Task"],
        the_hook: typing.Union[None, hook.BaseHook],
        the_ratelimiter: typing.Union[None, ratelimiter.BaseRateLimiter],
        loglevel: str,
        log_metadata: typing.Union[None, dict],
        log_handler: typing.Union[None, logger.BaseLogger],
    ) -> None:
        if not isinstance(the_broker, (type(None), broker.BaseBroker)):
            raise ValueError(
                """the_broker should be of type:: None or wiji.broker.BaseBroker You entered: {0}""".format(
                    type(the_broker)
                )
            )
        if not isinstance(queue_name, str):
            raise ValueError(
                """`queue_name` should be of type:: `str` You entered: {0}""".format(
                    type(queue_name)
                )
            )

        if not isinstance(task_name, (type(None), str)):
            raise ValueError(
                """`task_name` should be of type:: `None` or `str` You entered: {0}""".format(
                    type(task_name)
                )
            )
        if not isinstance(chain, (type(None), Task)):
            raise ValueError(
                """`chain` should be of type:: `None` or `wiji.task.Task` You entered: {0}""".format(
                    type(chain)
                )
            )
        if not isinstance(the_hook, (type(None), hook.BaseHook)):
            raise ValueError(
                """`the_hook` should be of type:: `None` or `wiji.hook.BaseHook` You entered: {0}""".format(
                    type(the_hook)
                )
            )
        if not isinstance(the_ratelimiter, (type(None), ratelimiter.BaseRateLimiter)):
            raise ValueError(
                """`the_ratelimiter` should be of type:: `None` or `wiji.ratelimiter.BaseRateLimiter` You entered: {0}""".format(
                    type(the_ratelimiter)
                )
            )
        if not isinstance(log_handler, (type(None), logger.BaseLogger)):
            raise ValueError(
                """`log_handler` should be of type:: `None` or `wiji.logger.BaseLogger` You entered: {0}""".format(
                    type(log_handler)
                )
            )
        if loglevel.upper() not in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
            raise ValueError(
                """`loglevel` should be one of; 'DEBUG', 'INFO', 'WARNING', 'ERROR' or 'CRITICAL'. You entered: {0}""".format(
                    loglevel
                )
            )
        if not isinstance(log_metadata, (type(None), dict)):
            raise ValueError(
                """`log_metadata` should be of type:: `None` or `dict` You entered: {0}""".format(
                    type(log_metadata)
                )
            )

        if not asyncio.iscoroutinefunction(self.run):
            raise ValueError(
                "The method: `run` of a class derived from: `wiji.task.Task` should be a python coroutine."
                "\nHint: did you forget to define the method using `async def` syntax?"
            )
        if not inspect.iscoroutinefunction(self.run):
            raise ValueError(
                "The method: `run` of a class derived from: `wiji.task.Task` should be a python coroutine."
                "\nHint: did you forget to define the method using `async def` syntax?"
            )
        if not asyncio.iscoroutinefunction(self.delay):
            raise ValueError(
                "The method: `delay` of a class derived from: `wiji.task.Task` should be a python coroutine."
                "\nHint: did you forget to define the method using `async def` syntax?"
            )
        if not inspect.iscoroutinefunction(self.delay):
            raise ValueError(
                "The method: `delay` of a class derived from: `wiji.task.Task` should be a python coroutine."
                "\nHint: did you forget to define the method using `async def` syntax?"
            )
        if not asyncio.iscoroutinefunction(self.retry):
            raise ValueError(
                "The method: `retry` of a class derived from: `wiji.task.Task` should be a python coroutine."
                "\nHint: did you forget to define the method using `async def` syntax?"
            )
        if not inspect.iscoroutinefunction(self.retry):
            raise ValueError(
                "The method: `retry` of a class derived from: `wiji.task.Task` should be a python coroutine."
                "\nHint: did you forget to define the method using `async def` syntax?"
            )

    def _sanity_check_logger(self, event: str) -> None:
        """
        Called when we want to make sure the supplied logger can log.
        This usually happens when we are instantiating a wiji.Task or a wiji.Worker
        """
        try:
            self.logger.log(logging.DEBUG, {"event": event})
        except Exception as e:
            raise e

    def _log(self, level: typing.Union[str, int], log_data: dict) -> None:
        # if the supplied logger is unable to log; we move on
        try:
            self.logger.log(level, log_data)
        except Exception:
            pass

    async def _broker_check(self, from_worker: bool) -> None:
        try:
            await self.the_broker.check(queue_name=self.queue_name)
            if not from_worker:
                self._checked_broker = True
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "wiji.Task.delay",
                    "stage": "end",
                    "state": "check broker failed",
                    "error": str(e),
                },
            )
            # exit with error
            raise ValueError(
                "The broker for task: `{0}` failed check request.".format(self.task_name)
            ) from e

    async def _notify_hook(
        self,
        state: TaskState,
        hook_metadata: str,
        execution_duration: typing.Union[None, typing.Dict[str, float]] = None,
        execution_exception: typing.Union[None, Exception] = None,
        return_value: typing.Union[None, typing.Any] = None,
    ) -> None:
        try:
            await self.the_hook.notify(
                task_name=self.task_name,
                task_id=self.task_options.task_id,
                queue_name=self.queue_name,
                state=state,
                hook_metadata=hook_metadata,
                execution_duration=execution_duration,
                execution_exception=execution_exception,
                return_value=return_value,
            )
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "wiji.Task._notify_hook",
                    "stage": "end",
                    "state": "task hook error",
                    "error": str(e),
                },
            )

    @abc.abstractmethod
    async def run(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        raise NotImplementedError("`run` method must be implemented.")

    async def delay(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """
        Parameters:
            args: The positional arguments to pass on to the task.
            kwargs: The keyword arguments to pass on to the task.
        """
        args, kwargs = self._validate_delay_args(*args, **kwargs)
        self._type_check(self.run, *args, **kwargs)
        if not self._checked_broker:
            await self._broker_check(from_worker=False)

        # every invocation of `my_task.delay()` is counted as unique and
        # should have a unique task_id even if it is a retry of a previous request
        self.task_options.task_id = str(uuid.uuid4())
        proto = protocol.Protocol(
            version=1,
            task_id=self.task_options.task_id,
            eta=self.task_options.eta,
            current_retries=self.task_options.current_retries,
            max_retries=self.task_options.max_retries,
            log_id=self.task_options.log_id,
            hook_metadata=self.task_options.hook_metadata,
            argsy=self.task_options.args,
            kwargsy=self.task_options.kwargs,
        )
        await self._notify_hook(
            state=TaskState.QUEUEING, hook_metadata=self.task_options.hook_metadata
        )
        try:
            await self.the_broker.enqueue(
                item=proto.json(), queue_name=self.queue_name, task_options=self.task_options
            )
            # this cannot raise an error since the method handles that error
            await self._notify_hook(
                state=TaskState.QUEUED, hook_metadata=self.task_options.hook_metadata
            )
        except TypeError as e:
            self._log(logging.ERROR, {"event": "wiji.Task.delay", "stage": "end", "error": str(e)})
            raise TypeError(
                "All the task arguments passed into `delay` should be JSON serializable."
            ) from e
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "wiji.Task.delay",
                    "stage": "end",
                    "state": "task queueing error",
                    "error": str(e),
                },
            )
            raise TaskDelayError("publishing to the broker failed.") from e

    def synchronous_delay(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.get_event_loop()

        loop.run_until_complete(self.delay(*args, **kwargs))

    async def retry(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """
        Parameters:
            args: The positional arguments to pass on to the task.
            kwargs: The keyword arguments to pass on to the task.

        Raises:
            WijiMaxRetriesExceededError: the task exceeded its max_retries count.
            WijiRetryError: the task is been retried. User applications should not capture this Exception.

        This method takes the same parameters as the `delay` method.
        It also behaves the same as `delay`
        """
        args, kwargs = self._validate_delay_args(*args, **kwargs)

        if self.task_options.current_retries >= self.task_options.max_retries:
            raise WijiMaxRetriesExceededError(
                "The task:`{task_name}` has reached its max_retries count of: {max_retries}".format(
                    task_name=self.task_name, max_retries=self.task_options.max_retries
                )
            )

        self.task_options.current_retries += 1
        await self.delay(*args, **kwargs)

        raise WijiRetryError(
            "Task: `{task_name}` is been retried. User applications should not capture this Exception!".format(
                task_name=self.task_name
            )
        )

    def _validate_delay_args(
        self, *args: typing.Any, **kwargs: typing.Any
    ) -> typing.Tuple[typing.Any, typing.Any]:
        for a in args:
            if isinstance(a, TaskOptions):
                raise ValueError(
                    """You cannot use a value of type `wiji.task.TaskOptions` as a normal argument.
                    \nHint: instead, pass it in as a kwarg(named argument)"""
                )
        for k, v in list(kwargs.items()):
            if isinstance(v, TaskOptions):
                self.task_options = v
                kwargs.pop(k)

        self.task_options.args = args
        self.task_options.kwargs = kwargs
        return self.task_options.args, self.task_options.kwargs

    @staticmethod
    def _type_check(func, *args: typing.Any, **kwargs: typing.Any) -> inspect.BoundArguments:
        """
        Check that `delay` is called with right arguments/signature.
        ie, the right arguments for the user implemented `run` method.

        if you have a func like:
            def foo(a, b, *args, c, d=10, **kwargs):
                pass
        you can type-check like:
            Task._type_check(foo, 1, 4)
        """
        sig = inspect.signature(func)
        return sig.bind(*args, **kwargs)


class _watchdogTask(Task):
    """
    This is a task that runs in the MainThread(as every other task).
    Its job is to start a new thread(Thread-<wiji_watchdog>) and communicate with it.
    That new thread will log a stack-trace if it detects any blocking calls(IO-bound, CPU-bound or otherwise) running on the MainThread.
    That trace is meant to help users of `wiji` be able to fix their applications.

    This task is always scheduled in the in-memory broker(`wiji.broker.InMemoryBroker`).
    """

    queue_name: str = "__WatchDogTaskQueue__"

    async def run(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        self._log(
            logging.DEBUG,
            {
                "event": "wiji.WatchDogTask.run",
                "state": "watchdog_run",
                "task_name": self.task_name,
                "task_id": self.task_options.task_id,
            },
        )
        await asyncio.sleep(0.1 / 1.5)


WatchDogTask = _watchdogTask(
    the_broker=broker.InMemoryBroker(), queue_name=_watchdogTask.queue_name, loglevel="WARNING"
)



import os
import sys
import string
import random
import asyncio
import logging
import argparse

import wiji
from cli import utils


os.environ["PYTHONASYNCIODEBUG"] = "1"


def make_parser() -> argparse.ArgumentParser:
    """
    this is abstracted into its own method so that it is easier to test it.
    """
    parser = argparse.ArgumentParser(
        prog="wiji",
        description="""wiji is an async distributed task queue.
                example usage:
                wiji-cli \
                --config dotted.path.to.a.wiji.conf.WijiConf.class.instance
                """,
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=wiji.__version__.about["__version__"]),
        help="The currently installed wiji version.",
    )
    parser.add_argument(
        "--config",
        required=True,
        help="The config file to use. \
        eg: --config dotted.path.to.a.wiji.conf.WijiConf.class.instance",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        required=False,
        default=False,
        help="""Whether we want to do a dry-run of the wiji cli.
        This is typically only used by developers who are developing wiji.
        eg: --dry-run""",
    )
    return parser


def main():
    """
    run as:
        wiji-cli --config dotted.path.to.a.wiji.conf.WijiConf.class.instance
    """
    worker_id = "".join(random.choices(string.ascii_uppercase + string.digits, k=17))
    logger = wiji.logger.SimpleLogger("wiji.cli")
    logger.log(logging.INFO, {"event": "wiji.cli.main", "stage": "start", "worker_id": worker_id})
    try:
        parser = make_parser()
        args = parser.parse_args()

        config = args.config
        dry_run = args.dry_run
        if dry_run:
            logger.log(
                logging.WARNING,
                "\n\n\t {} \n\n".format(
                    "Wiji: Caution; You have activated dry-run, wiji may not behave correctly."
                ),
            )

        config_instance = utils.load.load_class(config)
        if not isinstance(config_instance, wiji.conf.WijiConf):
            err = ValueError(
                """`config_instance` should be of type:: `wiji.conf.WijiConf` You entered: {0}""".format(
                    type(config_instance)
                )
            )
            logger.log(logging.ERROR, {"event": "wiji.cli.main", "stage": "end", "error": str(err)})
            sys.exit(77)

        if dry_run:
            logger.log(
                logging.INFO, {"event": "wiji.cli.main", "stage": "end", "state": "dry_run end"}
            )
            return

        asyncio_debug = False
        if os.environ.get("WIJI_DEBUG", None):
            asyncio_debug = True
        asyncio.run(async_main(logger=logger, config_instance=config_instance), debug=asyncio_debug)
    except Exception as e:
        logger.log(logging.ERROR, {"event": "wiji.cli.main", "stage": "end", "error": str(e)})
        sys.exit(77)
    finally:
        logger.log(logging.INFO, {"event": "wiji.cli.main", "stage": "end"})


async def async_main(logger: wiji.logger.BaseLogger, config_instance: wiji.conf.WijiConf) -> None:
    """
    (i)   set signal handlers.
    (ii)  consume tasks.
    (iii) continuously produce watchdog tasks.
    """
    watchdog_worker = wiji.Worker(
        the_task=wiji.task.WatchDogTask,
        use_watchdog=True,
        watchdog_duration=config_instance.watchdog_duration,
    )
    workers = [watchdog_worker]
    watch_dog_producer = [utils._producer.produce_tasks_continously(task=wiji.task.WatchDogTask)]

    for task in config_instance.tasks:
        _worker = wiji.Worker(the_task=task)
        workers.append(_worker)

    consumers = []
    for i in workers:
        consumers.append(i.consume_tasks())

    gather_tasks = asyncio.gather(
        *consumers, *watch_dog_producer, utils.sig._signal_handling(logger=logger, workers=workers)
    )
    await gather_tasks


if __name__ == "__main__":
    main()



import signal
import asyncio
import logging
import functools

import wiji


async def _signal_handling(logger: wiji.logger.BaseLogger, workers: list) -> None:
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.get_event_loop()

    try:
        for _signal in [signal.SIGHUP, signal.SIGINT, signal.SIGQUIT, signal.SIGTERM]:
            loop.add_signal_handler(
                _signal,
                functools.partial(
                    asyncio.ensure_future,
                    _handle_termination_signal(logger=logger, _signal=_signal, workers=workers),
                ),
            )
    except ValueError as e:
        logger.log(
            logging.DEBUG,
            {
                "event": "wiji.cli.signals",
                "stage": "end",
                "state": "this OS does not support the said signal",
                "error": str(e),
            },
        )


async def _handle_termination_signal(
    logger: wiji.logger.BaseLogger, _signal: int, workers: list
) -> None:
    logger.log(
        logging.INFO,
        {
            "event": "wiji.cli.signals",
            "stage": "start",
            "state": "received termination signal",
            "signal": _signal,
        },
    )

    shutdown_tasks = []
    for worker in workers:
        shutdown_tasks.append(worker.shutdown())
    # the shutdown process for all tasks needs to happen concurrently
    tasks = asyncio.gather(*shutdown_tasks)
    asyncio.ensure_future(tasks)

    def wait_worker_shutdown(workers) -> bool:
        no_workers = len(workers)
        success_shut_down = 0
        for worker in workers:
            if worker.SUCCESFULLY_SHUT_DOWN:
                success_shut_down += 1

        if success_shut_down == no_workers:
            loop_continue = False
        else:
            loop_continue = True
        return loop_continue

    while wait_worker_shutdown(workers=workers):
        logger.log(
            logging.INFO,
            {
                "event": "wiji.cli.signals",
                "stage": "start",
                "state": "waiting for all workers to drain properly",
            },
        )
        await asyncio.sleep(5)

    logger.log(
        logging.INFO,
        {
            "event": "wiji.cli.signals",
            "stage": "end",
            "state": "all workers have succesfully shutdown",
        },
    )
    return



import abc
import typing
import logging


class BaseLogger(abc.ABC):
    """
    """

    @abc.abstractmethod
    def bind(self, level: typing.Union[str, int], log_metadata: dict) -> None:
        """
        called when a wiji worker is been instantiated so that the logger can be
        notified of level & log_metadata that a user supplied to a wiji worker.
        The logger can choose to bind these log_metadata to itself.

        Parameters:
            level: logging level eg DEBUG
            log_metadata: log metadata that can be included in all log statements
        """
        raise NotImplementedError("`bind` method must be implemented.")

    @abc.abstractmethod
    def log(self, level: typing.Union[str, int], log_data: dict) -> None:
        """
        called by wiji everytime it wants to log something.

        Parameters:
            level: logging level eg `logging.INFO`
            log_data: the message to log
        """
        raise NotImplementedError("`log` method must be implemented.")


class SimpleLogger(BaseLogger):
    """
    This is an implementation of BaseLogger.
    It implements a structured logger that renders logs in a json/dict like manner.

    example usage:

    .. code-block:: python

        logger = SimpleLogger("myLogger")
        logger.bind(level="INFO",
                    log_metadata={"customer_id": "34541"})
        logger.log(logging.INFO,
                   {"event": "web_request", "url": "https://www.google.com/"})
    """

    def __init__(self, logger_name: str):
        """
        Parameters:
            logger_name: name of the logger. it should be unique per logger.
        """
        self.logger_name = logger_name
        self.logger: typing.Any = None

    def bind(self, level: typing.Union[str, int], log_metadata: dict) -> None:
        level = self._nameToLevel(level=level)

        self._logger = logging.getLogger(self.logger_name)
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)
        handler.setLevel(level)
        if not self._logger.handlers:
            self._logger.addHandler(handler)
        self._logger.setLevel(level)
        self.logger: logging.LoggerAdapter = wijiLoggingAdapter(self._logger, log_metadata)

    def log(self, level: typing.Union[str, int], log_data: dict) -> None:
        level = self._nameToLevel(level=level)

        if not self.logger:
            self.bind(level=level, log_metadata={})
        if level >= logging.ERROR:
            self.logger.log(level, log_data, exc_info=True)
        else:
            self.logger.log(level, log_data)

    def _nameToLevel(self, level: typing.Union[str, int]) -> int:
        try:
            if isinstance(level, str):
                # please mypy
                _level: int = logging._nameToLevel[level.upper()]
            else:
                _level = level
            return _level
        except KeyError as e:
            raise ValueError(
                "`level` should be one of; 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'FATAL', or 'CRITICAL'"
            ) from e


class wijiLoggingAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        if isinstance(msg, str):
            merged_msg = "{0} {1}".format(msg, self.extra)
            if self.extra == {}:
                merged_msg = "{0}".format(msg)
            return merged_msg, kwargs
        else:
            merged_msg = {**msg, **self.extra}
            return "{0}".format(merged_msg), kwargs



import json
import typing
import datetime


class Protocol:
    def __init__(
        self,
        version: int,
        task_id: str,
        eta: typing.Union[float, str],
        current_retries: int,
        max_retries: int,
        log_id: str,
        hook_metadata: str,
        argsy: tuple,
        kwargsy: dict,
    ):
        self._validate_protocol_args(
            version=version,
            task_id=task_id,
            eta=eta,
            current_retries=current_retries,
            max_retries=max_retries,
            log_id=log_id,
            hook_metadata=hook_metadata,
            argsy=argsy,
            kwargsy=kwargsy,
        )

        self.version = version
        self.task_id = task_id
        self.current_retries = current_retries
        self.max_retries = max_retries
        self.log_id = log_id
        self.hook_metadata = hook_metadata
        self.argsy = argsy
        self.kwargsy = kwargsy

        if isinstance(eta, float):
            self.eta = self._eta_to_isoformat(eta=eta)
        else:
            self.eta = eta

    def _validate_protocol_args(
        self,
        version,
        task_id,
        eta,
        current_retries,
        max_retries,
        log_id,
        hook_metadata,
        argsy,
        kwargsy,
    ):
        if not isinstance(version, int):
            raise ValueError(
                """`version` should be of type:: `int` You entered: {0}""".format(type(version))
            )
        if not isinstance(task_id, str):
            raise ValueError(
                """`task_id` should be of type:: `str` You entered: {0}""".format(type(task_id))
            )
        if not isinstance(eta, (str, float)):
            raise ValueError(
                """`eta` should be of type:: `str` or `float` You entered: {0}""".format(type(eta))
            )
        if isinstance(eta, str):
            try:
                # type-check that string eta is ISO 8601-formatted
                self._from_isoformat(eta)
            except Exception as e:
                raise ValueError(
                    """`eta` in string format should be a python ISO 8601-formatted string. You entered: {0}""".format(
                        eta
                    )
                ) from e
        if not isinstance(current_retries, int):
            raise ValueError(
                """`current_retries` should be of type:: `int` You entered: {0}""".format(
                    type(current_retries)
                )
            )
        if not isinstance(max_retries, int):
            raise ValueError(
                """`max_retries` should be of type:: `int` You entered: {0}""".format(
                    type(max_retries)
                )
            )
        if not isinstance(log_id, str):
            raise ValueError(
                """`log_id` should be of type:: `str` You entered: {0}""".format(type(log_id))
            )
        if not isinstance(hook_metadata, (type(None), str)):
            raise ValueError(
                """`hook_metadata` should be of type:: `None` or `str` You entered: {0}""".format(
                    type(hook_metadata)
                )
            )
        if not isinstance(argsy, tuple):
            raise ValueError(
                """`argsy` should be of type:: `tuple` You entered: {0}""".format(type(argsy))
            )
        if not isinstance(kwargsy, dict):
            raise ValueError(
                """`kwargsy` should be of type:: `dict` You entered: {0}""".format(type(kwargsy))
            )

    @staticmethod
    def _eta_to_isoformat(eta):
        """
        converts eta in float seconds to python's ISO 8601-formatted datetime.
        """
        eta = datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(seconds=eta)
        eta = eta.isoformat()
        return eta

    @staticmethod
    def _from_isoformat(eta):
        """
        converts ISO 8601-formatted datetime to python datetime
        usage:
            dt = Protocol._from_isoformat(eta="2020-04-01T18:41:54.195638+00:00")
        """
        dt = datetime.datetime.strptime(eta, "%Y-%m-%dT%H:%M:%S.%f%z")
        return dt

    def json(self):
        return json.dumps(
            {
                "version": self.version,
                "task_id": self.task_id,
                "eta": self.eta,
                "current_retries": self.current_retries,
                "max_retries": self.max_retries,
                "log_id": self.log_id,
                "hook_metadata": self.hook_metadata,
                "args": self.argsy,
                "kwargs": self.kwargsy,
            }
        )



from .worker import Worker  # noqa: F401

from . import task  # noqa: F401
from . import hook  # noqa: F401
from . import conf  # noqa: F401
from . import broker  # noqa: F401
from . import logger  # noqa: F401
from . import ratelimiter  # noqa: F401


from . import __version__  # noqa: F401



import typing

from . import task


class WijiConf:
    def __init__(self, tasks: typing.List["task.Task"], watchdog_duration: float = 0.1) -> None:
        self._validate_config_args(tasks=tasks, watchdog_duration=watchdog_duration)

        self.tasks = tasks
        self.watchdog_duration = watchdog_duration

    def _validate_config_args(self, tasks, watchdog_duration):
        if not isinstance(tasks, list):
            raise ValueError(
                """`tasks` should be of type:: `list` You entered: {0}""".format(type(tasks))
            )
        for tsk in tasks:
            if not isinstance(tsk, task.Task):
                raise ValueError(
                    """`task` should be of type:: `wiji.task.Task` You entered: {0}""".format(
                        type(tsk)
                    )
                )
        if not isinstance(watchdog_duration, float):
            raise ValueError(
                """`watchdog_duration` should be of type:: `float` You entered: {0}""".format(
                    type(watchdog_duration)
                )
            )



import abc
import typing
import logging

from . import logger

if typing.TYPE_CHECKING:
    from . import task


class BaseHook(abc.ABC):
    """
    """

    @abc.abstractmethod
    async def notify(
        self,
        task_name: str,
        task_id: str,
        queue_name: str,
        hook_metadata: str,
        state: "task.TaskState",
        execution_duration: typing.Union[None, typing.Dict[str, float]] = None,
        execution_exception: typing.Union[None, Exception] = None,
        return_value: typing.Union[None, typing.Any] = None,
    ) -> None:
        """
        called by `wiji` worker whenever a task undergoes a state change.
        """
        raise NotImplementedError("`notify` method must be implemented.")


class SimpleHook(BaseHook):
    """
    This is an implementation of BaseHook.
    It implements a no-op hook that does nothing when called
    """

    def __init__(self, log_handler: typing.Union[None, logger.BaseLogger] = None) -> None:
        if log_handler is not None:
            self.logger = log_handler
        else:
            self.logger = logger.SimpleLogger("wiji.hook.SimpleHook")

    async def notify(
        self,
        task_name: str,
        task_id: str,
        queue_name: str,
        hook_metadata: str,
        state: "task.TaskState",
        execution_duration: typing.Union[None, typing.Dict[str, float]] = None,
        execution_exception: typing.Union[None, Exception] = None,
        return_value: typing.Union[None, typing.Any] = None,
    ) -> None:
        self.logger.log(
            logging.NOTSET,
            {
                "event": "wiji.SimpleHook.notify",
                "stage": "start",
                "state": state,
                "task_name": task_name,
                "task_id": task_id,
                "queue_name": queue_name,
                "hook_metadata": hook_metadata,
                "execution_duration": execution_duration,
                "execution_exception": str(execution_exception),
                "return_value": str(return_value),
            },
        )



import abc
import time
import typing
import asyncio
import logging

from . import task
from . import logger


# TODO: rate limiting should take into account number of succesful task executions(based on whether they raised an exception or not)
# and also the number of failures.
# We should also pass in the task_name, and exception raised
# This will enable users of wiji to come up with better rate limiting algos.
# So: we need to make the BaseRateLimiter accept this more arguments.
# We also need to make SimpleRateLimiter a bit smarter and take this new args in effect(or maybe not[probably NOT])


class BaseRateLimiter(abc.ABC):
    """
    This is the interface that must be implemented to satisfy wiji's rate limiting.
    User implementations should inherit this class and
    implement the :func:`limit <BaseRateLimiter.limit>` methods with the type signatures shown.

    It may be important to control the rate at which the worker(wiji) consumes/executes tasks.
    wiji lets you do this, by allowing you to specify a custom rate limiter.
    """

    @abc.abstractmethod
    async def limit(self) -> None:
        """
        rate limit consumation/execution of tasks.
        """
        raise NotImplementedError("`limit` method must be implemented.")

    @abc.abstractmethod
    async def execution_outcome(
        self,
        task_name: str,
        task_id: str,
        queue_name: str,
        execution_duration: typing.Dict[str, float],
        execution_exception: typing.Union[None, Exception],
        return_value: typing.Any,
    ) -> None:
        """
        this method is called by the worker once it has finished executing a task.
        implementers may choose to use the metrics provided to dynamically adjust their rate limiting policies.
        """
        raise NotImplementedError("`execution_outcome` method must be implemented.")


class SimpleRateLimiter(BaseRateLimiter):
    """
    This is an implementation of BaseRateLimiter.

    It does rate limiting using a `token bucket rate limiting algorithm <https://en.wikipedia.org/wiki/Token_bucket>`_

    example usage:

    .. code-block:: python

        rateLimiter = SimpleRateLimiter(log_handler=myLogger, execution_rate=10.0)
        await rateLimiter.limit()
    """

    def __init__(
        self,
        execution_rate: float = 100_000_000.0,
        log_handler: typing.Union[None, logger.BaseLogger] = None,
    ) -> None:
        """
        Parameters:
            execution_rate: the maximum rate, in tasks/second, at which wiji can consume/execute tasks.
        """
        self._validate_args(execution_rate=execution_rate, log_handler=log_handler)

        self.execution_rate: float = execution_rate
        self.max_tokens: float = self.execution_rate
        self.tokens: float = self.max_tokens

        self.delay_for_tokens: float = 1.0
        self.updated_at: float = time.monotonic()
        self.tasks_executed: int = 0
        self.effective_execution_rate: float = 0.0

        if log_handler is not None:
            self.logger = log_handler
        else:
            self.logger = logger.SimpleLogger("wiji.ratelimiter.SimpleRateLimiter")

    def _validate_args(self, execution_rate, log_handler):
        if not isinstance(execution_rate, float):
            raise ValueError(
                """`execution_rate` should be of type:: `float` You entered: {0}""".format(
                    type(execution_rate)
                )
            )
        if execution_rate < 1.0:
            raise ValueError(
                """`execution_rate` should not be less than 1.0 You entered: {0}""".format(
                    execution_rate
                )
            )
        if not isinstance(log_handler, (type(None), logger.BaseLogger)):
            raise ValueError(
                """`log_handler` should be of type:: `None` or `wiji.logger.BaseLogger` You entered: {0}""".format(
                    type(log_handler)
                )
            )

    def _add_new_tokens(self) -> None:
        now = time.monotonic()
        time_since_update = now - self.updated_at
        self.effective_execution_rate = self.tasks_executed / time_since_update
        new_tokens = time_since_update * self.execution_rate
        if new_tokens > 1:
            self.tokens = min(self.tokens + new_tokens, self.max_tokens)
            self.updated_at = now
            self.tasks_executed = 0

    async def limit(self) -> None:
        self.logger.log(logging.INFO, {"event": "wiji.SimpleRateLimiter.limit", "stage": "start"})
        while self.tokens < 1:
            self._add_new_tokens()
            # todo: sleep in an exponetial manner upto a maximum then wrap around.
            await asyncio.sleep(self.delay_for_tokens)
            self.logger.log(
                logging.INFO,
                {
                    "event": "wiji.SimpleRateLimiter.limit",
                    "stage": "end",
                    "state": "limiting rate",
                    "execution_rate": self.execution_rate,
                    "delay": self.delay_for_tokens,
                    "effective_execution_rate": self.effective_execution_rate,
                },
            )
        self.tasks_executed += 1
        self.tokens -= 1

    async def execution_outcome(
        self,
        task_name: str,
        task_id: str,
        queue_name: str,
        execution_duration: typing.Dict[str, float],
        execution_exception: typing.Union[None, Exception],
        return_value: typing.Any,
    ) -> None:
        """
        SimpleRateLimiter does nothing with the data/metrics it gets about execution outcome.
        However, you can imagine a smarter RateLimiter that uses these metrics to dynamically change
        its rate-limiting methodologies; eg increase ratelimit if percentage of exceptions goes up.
        """
        if queue_name != task._watchdogTask.queue_name:
            self.logger.log(
                logging.DEBUG,
                {
                    "event": "wiji.SimpleRateLimiter.execution_outcome",
                    "stage": "end",
                    "state": "execution_outcome record",
                    "task_name": task_name,
                    "task_id": task_id,
                    "queue_name": queue_name,
                    "execution_duration": execution_duration,
                    "execution_exception": str(execution_exception),
                    "return_value": str(return_value),
                },
            )



import abc
import typing
import asyncio

if typing.TYPE_CHECKING:
    from . import task


class BaseBroker(abc.ABC):
    """
    This is the interface that must be implemented to satisfy wiji's broker.
    User implementations should inherit this class and
    implement the :func:`check <BaseBroker.check>`, :func:`enqueue <BaseBroker.enqueue>` and
    :func:`dequeue <BaseBroker.dequeue>` methods with the type signatures shown.

    wiji calls an implementation of this class to enqueue and/or dequeue an item.
    """

    @abc.abstractmethod
    async def check(self, queue_name: str) -> None:
        """
        called by `wiji` worker once, during startup so as;
          - to check that the broker is up
          - to inform the broker of the queue_name that the worker will be consuming from.
            the broker can go ahead and create this queue_name if it does not exist
        """
        raise NotImplementedError("`check` method must be implemented.")

    @abc.abstractmethod
    async def enqueue(self, item: str, queue_name: str, task_options: "task.TaskOptions") -> None:
        """
        enqueue/save an item.

        Parameters:
            item: The item to be enqueued/saved
            queue_name: name of queue to enqueue in
            task_options: options for the specific task been enqueued
        """
        raise NotImplementedError("`enqueue` method must be implemented.")

    @abc.abstractmethod
    async def dequeue(self, queue_name: str) -> str:
        """
        dequeue an item.

        Returns:
            item that was dequeued
        """
        raise NotImplementedError("`dequeue` method must be implemented.")

    @abc.abstractmethod
    async def done(
        self, item: str, queue_name: str, task_options: "task.TaskOptions", state: "task.TaskState"
    ) -> None:
        """
        called by wiji worker once it is done executing a task.
        the broker can then decide to do any clean up actions like removing that task from the queue etc.
        """
        raise NotImplementedError("`done` method must be implemented.")

    @abc.abstractmethod
    async def shutdown(self, queue_name: str, duration: float) -> None:
        """
        called by wiji worker when it receives a shutdown signal like `SIGTERM`.
        the broker can decide to perform shutdown procedures like releasing connections/file descriptors etc.

        Parameters:
            queue_name: name of queue which the wiji worker was consuming from
            duration: duration in seconds that wiji worker will wait for after calling this method.
        """
        raise NotImplementedError("`shutdown` method must be implemented.")


class InMemoryBroker(BaseBroker):
    """
    This broker should only be used for:
      (i) tests
      (ii) demos
      (iii) the watchdog task.
    Do not use this broker in production or anywhere else that you care about.

    {
        "queue1": ["item1", "item2", "item3"],
        "queue2": ["item1", "item2", "item3"]
        ...
    }
    """

    def __init__(self) -> None:
        """
        """
        self.store: dict = {}

    async def check(self, queue_name: str) -> None:
        if queue_name not in self.store:
            self.store[queue_name] = []
        await asyncio.sleep(1 / 117)

    async def enqueue(self, item: str, queue_name: str, task_options: "task.TaskOptions") -> None:
        if self.store.get(queue_name):
            self.store[queue_name].append(item)
            await asyncio.sleep(delay=-1)
        else:
            self.store[queue_name] = [item]
            await asyncio.sleep(delay=-1)

    async def dequeue(self, queue_name: str) -> str:
        while True:
            if queue_name in self.store:
                try:
                    return await asyncio.sleep(delay=-1, result=self.store[queue_name].pop(0))
                except IndexError:
                    # queue is empty
                    await asyncio.sleep(5)
            else:
                raise ValueError("queue with name: {0} does not exist.".format(queue_name))

    async def done(
        self, item: str, queue_name: str, task_options: "task.TaskOptions", state: "task.TaskState"
    ) -> None:
        """
        for this broker, this method is not needed, since `dequeue` uses .pop() which deletes the item.
        """
        if queue_name in self.store:
            return await asyncio.sleep(delay=-1, result=None)
        else:
            raise ValueError("queue with name: {0} does not exist.".format(queue_name))

    async def shutdown(self, queue_name: str, duration: float) -> None:
        return await asyncio.sleep(delay=-1, result=None)



import typing
import asyncio

import wiji


async def produce_tasks_continously(
    task: wiji.task.Task, *args: typing.Any, **kwargs: typing.Any
) -> None:
    while True:
        await task.delay(*args, **kwargs)
        await asyncio.sleep(1 / 117)



import os
import json
import time
import random
import string
import typing
import logging
import asyncio
import datetime

from . import task
from . import protocol
from . import watchdog


class Worker:
    """
    The only time this worker coroutine should ever raise an Exception is either:
      - during class instantiation
      - when the worker is about to start consuming tasks
    """

    def __init__(
        self,
        the_task: task.Task,
        worker_id: typing.Union[None, str] = None,
        use_watchdog: bool = False,
        watchdog_duration: float = 0.1,
    ) -> None:
        """
        """
        self._validate_worker_args(
            the_task=the_task,
            worker_id=worker_id,
            use_watchdog=use_watchdog,
            watchdog_duration=watchdog_duration,
        )

        self._PID = os.getpid()
        self.the_task = the_task
        if worker_id is not None:
            self.worker_id = worker_id
        else:
            self.worker_id = "".join(random.choices(string.ascii_uppercase + string.digits, k=17))

        self.the_task.log_metadata.update({"worker_id": self.worker_id, "process_id": self._PID})
        self.the_task.logger.bind(
            level=self.the_task.loglevel, log_metadata=self.the_task.log_metadata
        )

        self.use_watchdog = use_watchdog
        self.watchdog_duration = watchdog_duration

        self.watchdog = None
        if self.use_watchdog:
            self.watchdog = watchdog.BlockingWatchdog(
                watchdog_duration=self.watchdog_duration, task_name=self.the_task.task_name
            )

        self.SHOULD_SHUT_DOWN: bool = False
        self.SUCCESFULLY_SHUT_DOWN: bool = False

        self.the_task._sanity_check_logger(event="worker_sanity_check_logger")

    def _validate_worker_args(
        self,
        the_task: task.Task,
        worker_id: typing.Union[None, str],
        use_watchdog: bool,
        watchdog_duration: float,
    ) -> None:
        if not isinstance(the_task, task.Task):
            raise ValueError(
                """`the_task` should be of type:: `wiji.task.Task` You entered: {0}""".format(
                    type(the_task)
                )
            )
        if not isinstance(worker_id, (type(None), str)):
            raise ValueError(
                """`worker_id` should be of type:: `None` or `str` You entered: {0}""".format(
                    type(worker_id)
                )
            )
        if not isinstance(use_watchdog, bool):
            raise ValueError(
                """`use_watchdog` should be of type:: `bool` You entered: {0}""".format(
                    type(use_watchdog)
                )
            )
        if not isinstance(watchdog_duration, float):
            raise ValueError(
                """`watchdog_duration` should be of type:: `float` You entered: {0}""".format(
                    type(watchdog_duration)
                )
            )

    def _log(self, level: typing.Union[str, int], log_data: dict) -> None:
        try:
            self.the_task.logger.log(level, log_data)
        except Exception:
            pass

    @staticmethod
    def _retry_after(current_retries: int) -> int:
        """
        returns the number of seconds to retry after.
        retries will happen in this sequence;
        0.5min, 1min, 2min, 4min, 8min, 16min, 32min, 16min, 16min, 16min ...
        """
        # TODO:
        # 1. give users ability to bring their own retry algorithms.
        if current_retries < 0:
            current_retries = 0

        jitter = random.randint(60, 180)  # 1min-3min
        if current_retries in [0, 1]:
            return int(0.5 * 60)  # 0.5min
        elif current_retries == 2:
            return 1 * 60
        elif current_retries >= 6:
            return (16 * 60) + jitter  # 16 minutes + jitter
        else:
            return (60 * (2 ** current_retries)) + jitter

    async def _notify_ratelimiter(
        self,
        return_value: typing.Any,
        execution_duration: typing.Dict[str, float],
        execution_exception: typing.Union[None, Exception],
    ) -> None:
        try:
            await self.the_task.the_ratelimiter.execution_outcome(
                task_name=self.the_task.task_name,
                task_id=self.the_task.task_options.task_id,
                queue_name=self.the_task.queue_name,
                execution_duration=execution_duration,
                execution_exception=execution_exception,
                return_value=return_value,
            )
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "wiji.Worker.run_task",
                    "stage": "end",
                    "state": "the_ratelimiter execution_outcome error",
                    "error": str(e),
                },
            )

    async def _notify_broker(
        self, item: str, queue_name: str, task_options: task.TaskOptions, state: task.TaskState
    ) -> None:
        try:
            await self.the_task.the_broker.done(
                item=item, queue_name=queue_name, task_options=task_options, state=state
            )
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "wiji.Worker.consume_tasks",
                    "stage": "end",
                    "state": "broker done error",
                    "error": str(e),
                },
            )

    async def run_task(self, *task_args: typing.Any, **task_kwargs: typing.Any) -> None:
        await self.the_task._notify_hook(
            state=task.TaskState.EXECUTING, hook_metadata=self.the_task.task_options.hook_metadata
        )
        if self.watchdog is not None:
            self.watchdog.notify_alive_before()

        return_value = None
        execution_exception = None
        thread_time_start = time.thread_time()
        perf_counter_start = time.perf_counter()
        monotonic_start = time.monotonic()
        process_time_start = time.process_time()
        try:
            return_value = await self.the_task.run(*task_args, **task_kwargs)
            if self.the_task.chain:
                # enqueue the chained task using the return_value
                await self.the_task.chain.delay(return_value)
        except task.WijiRetryError as e:
            # task is been retried
            self._log(
                logging.INFO,
                {
                    "event": "wiji.Worker.run_task",
                    "state": str(e),
                    "stage": "end",
                    "task_name": self.the_task.task_name,
                    "current_retries": self.the_task.task_options.current_retries,
                    "max_retries": self.the_task.task_options.max_retries,
                },
            )
        except Exception as e:
            execution_exception = e
            self._log(
                logging.ERROR,
                {
                    "event": "wiji.Worker.run_task",
                    "stage": "end",
                    "state": "task execution error",
                    "error": str(e),
                },
            )
        finally:
            thread_time_end = time.thread_time()
            perf_counter_end = time.perf_counter()
            monotonic_end = time.monotonic()
            process_time_end = time.process_time()
            execution_duration = {
                "thread_time": float("{0:.4f}".format(thread_time_end - thread_time_start)),
                "perf_counter": float("{0:.4f}".format(perf_counter_end - perf_counter_start)),
                "monotonic": float("{0:.4f}".format(monotonic_end - monotonic_start)),
                "process_time": float("{0:.4f}".format(process_time_end - process_time_start)),
            }
            await self._notify_ratelimiter(
                return_value=return_value,
                execution_duration=execution_duration,
                execution_exception=execution_exception,
            )
            await self.the_task._notify_hook(
                state=task.TaskState.EXECUTED,
                hook_metadata=self.the_task.task_options.hook_metadata,
                execution_duration=execution_duration,
                execution_exception=execution_exception,
                return_value=return_value,
            )
            if self.watchdog is not None:
                self.watchdog.notify_alive_after()

    async def consume_tasks(
        self, TESTING: bool = False
    ) -> typing.Union[None, typing.Dict[str, typing.Any]]:
        """
        In loop; dequeues items from the :attr:`queue <Worker.queue>` and calls :func:`run <Worker.run>`.

        Parameters:
            TESTING: indicates whether this method is been called while running tests.
        """
        # this can exit with error
        await self.the_task._broker_check(from_worker=True)

        if self.watchdog is not None:
            self.watchdog.start()

        dequeue_retry_count = 0
        while True:
            self._log(logging.INFO, {"event": "wiji.Worker.consume_tasks", "stage": "start"})
            if self.SHOULD_SHUT_DOWN:
                self._log(
                    logging.INFO,
                    {
                        "event": "wiji.Worker.consume_tasks",
                        "stage": "end",
                        "state": "cleanly shutting down worker.",
                    },
                )
                return None

            try:
                # rate limit ourselves
                await self.the_task.the_ratelimiter.limit()
            except Exception as e:
                self._log(
                    logging.ERROR,
                    {
                        "event": "wiji.Worker.consume_tasks",
                        "stage": "end",
                        "state": "consume_tasks error",
                        "error": str(e),
                    },
                )
                continue

            try:
                _dequeued_item: str = await self.the_task.the_broker.dequeue(
                    queue_name=self.the_task.queue_name
                )
                dequeued_item: dict = json.loads(_dequeued_item)
            except Exception as e:
                poll_queue_interval = self._retry_after(dequeue_retry_count)
                dequeue_retry_count += 1
                self._log(
                    logging.ERROR,
                    {
                        "event": "wiji.Worker.consume_tasks",
                        "stage": "end",
                        "state": "dequeue tasks failed. sleeping for {0}minutes".format(
                            poll_queue_interval / 60
                        ),
                        "dequeue_retry_count": dequeue_retry_count,
                        "error": str(e),
                    },
                )
                await asyncio.sleep(poll_queue_interval)
                continue

            # dequeue succeded
            dequeue_retry_count = 0
            try:
                _ = dequeued_item["version"]
                task_id = dequeued_item["task_id"]
                task_eta = dequeued_item["eta"]
                _ = dequeued_item["current_retries"]
                _ = dequeued_item["max_retries"]
                task_log_id = dequeued_item["log_id"]
                task_hook_metadata = dequeued_item["hook_metadata"]
                task_args = dequeued_item["args"]
                task_kwargs = dequeued_item["kwargs"]
            except KeyError as e:
                e = KeyError("enqueued message/object is missing required field: {}".format(str(e)))
                self._log(
                    logging.ERROR,
                    {
                        "event": "wiji.Worker.consume_tasks",
                        "stage": "end",
                        "state": "consume_tasks error",
                        "error": str(e),
                    },
                )
                continue

            await self.the_task._notify_hook(
                state=task.TaskState.DEQUEUED, hook_metadata=task_hook_metadata
            )

            now = datetime.datetime.now(tz=datetime.timezone.utc)
            if protocol.Protocol._from_isoformat(task_eta) <= now:
                await self.run_task(*task_args, **task_kwargs)
                await self._notify_broker(
                    item=_dequeued_item,
                    queue_name=self.the_task.queue_name,
                    task_options=self.the_task.task_options,
                    state=task.TaskState.EXECUTED,
                )
            else:
                # respect eta
                await self.the_task.delay(*task_args, **task_kwargs)
            self._log(
                logging.INFO,
                {
                    "event": "wiji.Worker.consume_tasks",
                    "stage": "end",
                    "log_id": task_log_id,
                    "task_id": task_id,
                },
            )
            if TESTING:
                # offer escape hatch for tests to come out of endless loop
                return dequeued_item

    async def shutdown(self) -> None:
        """
        Cleanly shutdown this worker.
        """
        self._log(
            logging.INFO,
            {
                "event": "wiji.Worker.shutdown",
                "stage": "start",
                "state": "intiating shutdown",
                "drain_duration": self.the_task.task_options.drain_duration,
            },
        )
        self.SHOULD_SHUT_DOWN = True
        if self.watchdog is not None:
            self.watchdog.stop()

        # half spent waiting for the broker, the other half just sleeping
        wait_duration = self.the_task.task_options.drain_duration / 2
        try:
            # asyncio.wait takes a python set as a first argument
            # after expiration of timeout, asyncio.wait does not cancel the task;
            # thus the broker shutdown can still continue on its own if it can.
            await asyncio.wait(
                {
                    self.the_task.the_broker.shutdown(
                        queue_name=self.the_task.queue_name, duration=wait_duration
                    )
                },
                timeout=wait_duration,
            )
        except Exception as e:
            self._log(
                logging.ERROR,
                {
                    "event": "wiji.Worker.shutdown",
                    "stage": "end",
                    "state": "calling broker shutdown error",
                    "error": str(e),
                },
            )

        # sleep so that worker can finish executing any tasks it had already dequeued.
        # we need to use asyncio.sleep so that we do not block eventloop.
        # this way, we do not prevent any other workers in the same loop from also shutting down cleanly.
        await asyncio.sleep(wait_duration)
        self.SUCCESFULLY_SHUT_DOWN = True
